# SaverSwitcher (c) 2007 by Peter Watkins (peterw@tux.org) 
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, which should be included with this software.

package Plugins::SaverSwitcher::Plugin;

use strict;

use CGI;
use Plugins::SaverSwitcher::Logging;
use Plugins::SaverSwitcher::Public;
use Slim::Control::Command;
use Slim::Control::Request;
use Slim::Utils::DateTime;
use Slim::Utils::Strings qw (string);
use Slim::Utils::Misc qw( msg );
use POSIX qw(strftime);

# a constant string to identify this screensaver
my $screensaverName = 'SCREENSAVER.plugin_SaverSwitcher';

# create a logging object
my $log = Plugins::SaverSwitcher::Logging::new('SaverSwitcher');

# ------------------------- SETTINGS ----------------------------------

# Logging:
# For setTypes(), possible string names are 'all', 'menu', 'web', 'buttons', 'timers', 'prefs', 'funcs' 
#
# uncomment the next two lines to log menu-related and preference-related items in addition to d_plugin default messages:
#$log->setTypes('menu','prefs');
#$log->setLevel(100);
#
# For function logging:
#$log->setTypes('funcs');
#$log->setLevel(100);
#
# For button/IR logging:
#$log->setTypes('buttons');
#$log->setLevel(100);
#
# uncomment the next two lines to log everything:
#$log->setTypes('all');
#$log->setLevel(100);
#$log->setTypes('misc');
#$log->setLevel(60);

my $minActiveDuration = 5;	# smallest number of seconds allowed for a screensaver duration
my $lookInterval = 3.5;		# for When Playing, check every $lookInterval seconds for evidence of a new song
my $minSwitchInterval = 1.25;	# must be this many seconds since last saver switch for button switch to work

# default (fallback) savers that work on any player
my %defaultSavers = (
	'screensaver' => 'playlist',
	'idlesaver'   => 'SCREENSAVER.datetime',
	'offsaver'    => 'SCREENSAVER.datetime',
);
# screensavers that require certain hardware
# 0 = slimp3
# 1 = SB1 (text)
# 2 = SB1 (graphical)
# 3 = SB2/SB3/Transporter
my %saverMinimums = (
	"SCREENSAVER.snow" => 2,
	"SCREENSAVER.visualizer_analog_vumeter" => 3,
	"SCREENSAVER.visualizer_digital_vumeter" => 3,
	"SCREENSAVER.visualizer_spectrum" => 3,
);
# Screensavers that we are pretty sure don't work with SaverSwitcher
my @incompatibleSavers = ( 
	$screensaverName,	# never list ourselves as an option
	'nosaver',		# "None" saver -- not really a screensaver, results in "stuck" displays
	'screensaver',		# "Now Playing, jump back" -- use regular Now Playing instead
);
# ------------------------- SETTINGS ----------------------------------

# globals
my @saverNames;		# names used for preferences, e.g. 'SCREENSAVER.plugin_SaverSwitcher'
my @saverStringNames;	# keys for string() localization, e.g. 'PLUGIN_SAVERSWITCHER'
my @menuItems;		# up/down choice arrays
$menuItems[0] = ['SETUP_SCREENSAVER', 'SETUP_OFFSAVER', 'SETUP_IDLESAVER', 'PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL'];
$menuItems[1] = ['1', '2', '3', '4', '5', '6', '7', '8', '9'];
# $menuItems[2] doesn't exist -- x2 is for duration counts
# $menuItems[3] is the array of screensaver l10n strings
my @buttonChoices = ( 'DISABLED', 'ENABLED' );
my $originalIRCommand; # need to keep track of original IR function
my %preChangeFunctions;	# for registerPreUsageCheck() coderefs

# per-player
# for screensaver operation:
my %saverPos;		# keep track of cycle through the screensavers per-player+saverType
my %playerSaver;	# keep track of player+saverTypes that use this screensaver
my %multiMode;		# keep track of what mode we're serving for each player
my %expectedMode;	# helps track moving between savers
my %latestSong;		# for noticing when a new song starts
my %latestDisplayed;	# for noticing when a stream sends new metadata
my %lastChanged;	# don't change too often
my %manualOverride;	# for manual selection
my %lastIRcode;		# last $ircode seen from this player (help track held buttons)
my %firstIRtime;	# first time this $ircode seen (help track held buttons)
my %lastIRtime;		# last time this player sent an IR code (help track held buttons)
my %jumpButton;		# friendly name of code we're currently considering acting on
my %passCode;		# $ircode that has been held on a given player and should be passed to other routines
my %durationToUse;	# override normal duration
# for menu operation
my %xPos;		# menu left/right pos
my %yPos;		# menu up/down pos
my %xHash;		# breadcrumbs through menu hierarchy
my %menuSaverNpos;	# which of 1-N SaverSwitcher slots we're in in the menu hierarchy
my %line2Hash;		# allows for line1 "breadcrumbs" as users navigate right

# API versions:
# 1.0: original -- no functions exposed
# 1.1: BETA support for registerPreUsageCheck()
# 1.2: BETA support for registerPreUsageCheck() with named hash input/return values
# 2.0: support for registerPreUsageCheck() with named hash input/return values
$Plugins::SaverSwitcher::Plugin::apiVersion = 2.0;

use vars qw($VERSION);
$VERSION = &vlVersion();

# our i18n/l10n strings
sub strings() { 
  	return '
MANUAL_ADJECTIVE
	EN	Manual
	ES	Manual
	FR	Manuel
	DE	Manuell

PLUGIN_SAVERSWITCHER
	EN	SaverSwitcher

PLUGIN_SAVERSWITCHER_SAVER
	EN	SaverSwitcher screensavers

PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL
	EN	Button control
	DE	Knopf Steuerung
	ES	Control con botones
	FR	Commande avec boutons
'};

# whether we're enabled (the callback remains in the function 
# stack after "disabling" the plugin, so we need to keep track of this)
my $pluginEnabled = 0;

# whether we've already subscribed to new client events
my $callbackset = 0;

# setup routine
sub initPlugin {
	# note that we should act
	$pluginEnabled = 1;
	&loadScreensaverList();
	&initConfig();
	if ( $callbackset == 0 ) {
		$log->msg("registering callback\n");
		Slim::Control::Request::subscribe( \&newPlayerReset, [['client']],[['new']]);
		Slim::Control::Request::subscribe( \&newSong, [['playlist']],[['newsong']]);
                $log->msg("wrapping button functions\n");
                $originalIRCommand = Slim::Control::Request::addDispatch(['ir','_ircode','_time'],[1, 0, 0, \&SaverSwitcher_irCommand]);
                if ( (!defined($originalIRCommand)) || (ref($originalIRCommand ) ne 'CODE') ) {
                        $log->msg("problem wrapping IR command!\n");
                }
		$callbackset = 1;
	}
	&resetPlayerPrefs(undef,undef);
	$log->msg("enabled");
}

sub shutdownPlugin {
	# we should not act
	$pluginEnabled = 0;
	$log->msg("asked to disable");
}

sub enabled {
	if ( substr($::VERSION,0,3) gt '6.5' ) {
		&Plugin_msg("This plugin has not been tested with SlimServer newer than 6.5.x\n");
	}
	return ($::VERSION ge '6.5');
}

sub vlVersion() {
	my $rcsVersion = '$Revision: 1.16 $';
	$rcsVersion =~ s/.*:\s*([0-9\.]*).*$/$1/;
	return $rcsVersion;
}

sub getDisplayName() {
	return 'PLUGIN_SAVERSWITCHER_SAVER';
}

sub resetIRcounter {
	my ($client,$id) = @_;
	$log->msg('buttons',70,"resetting IR counters for $id");
	$lastIRcode{$id} = undef;
}

# our IR wrapper function
sub SaverSwitcher_irCommand {
        my @args = @_;
	my $buttonControl = Slim::Utils::Prefs::get("$screensaverName-button-control");
	if ( ($buttonControl == 0) || ($pluginEnabled == 0) ) {
		$log->msg('buttons',80,"button control disabled, calling original IR command");
        	return &$originalIRCommand(@args);
	}
        my $request = $args[0];
        my $ircode = $request->getParam('_ircode');
	my $client = $request->client();
	if (! defined($client) ) {
		$log->msg('buttons',80,"no client! calling original command");
        	return &$originalIRCommand(@args);
	}
	my $id = $client->id();
	my $codename = Slim::Hardware::IR::lookupCodeBytes($client,$ircode);
	my $clientMode = Slim::Buttons::Common::mode($client);
	$log->msg('buttons',90,"client $id mode $clientMode ircode $ircode codename $codename multiMode \"$multiMode{$id}\"");
	if ( $codename !~ m|^[0-9]$| ) {
		$log->msg('buttons',70,"not a digit, calling original IR command");
        	return &$originalIRCommand(@args);
	}
	if ( (!defined($lastIRcode{$id})) || ($lastIRcode{$id} ne $ircode) ) {
		$lastIRcode{$id} = $ircode;
		$firstIRtime{$id} = Time::HiRes::time();
		$passCode{$id} = undef;
	}
	$lastIRtime{$id} = Time::HiRes::time();
	$log->msg('timers',50,"$id: new song, killing resetIRCounter() timers");
	Slim::Utils::Timers::killTimers($client, \&resetIRcounter);
	Slim::Utils::Timers::setTimer($client,Time::HiRes::time()+(2 * $Slim::Hardware::IR::IRSINGLETIME),\&resetIRcounter,$id);
	if ( defined($passCode{$id}) && ($passCode{$id} eq $ircode) ) {
		# pass this along
		$log->msg('buttons',60,"this code has been held, calling original IR command");
		$jumpButton{$id} = undef;
        	return &$originalIRCommand(@args);
       	} elsif ( defined($multiMode{$id}) && ($clientMode !~ m|^INPUT\.|) ) {
		# SaverSwitcher is active, we might want to act
		if ( ($lastIRtime{$id} - $firstIRtime{$id}) >  $Slim::Hardware::IR::IRSINGLETIME ) {
			# button held
			if ( defined($jumpButton{$id}) && ($jumpButton{$id} eq $codename) ) {
				# cancel jump!
				$log->msg('buttons',50,"button held, will not jump to specified saver");
				$log->msg('timers',25,"$id: killing jumpToSaver() timers");
				Slim::Utils::Timers::killTimers($client, \&jumpToSaver);
				$jumpButton{$id} = undef;
			}
			# OK to call normal IR command next time
			$passCode{$id} = $ircode;
			# and call it now
			$log->msg('buttons',60,"this code has been held, calling original IR command");
        		return &$originalIRCommand(@args);
		}
		my $saverType = $multiMode{$id};
		my $idstring = "$id\t$saverType";
		my $pSaver = $playerSaver{$idstring};
		# decide whether to act              
		if ( defined($saverType) && defined($pSaver) && ($pSaver ne '') ) {
			my $act = 0;
			my $saverNumber = ($codename - 1);
			if ( $saverNumber < 0 ) { 
				$saverNumber = undef; 
				$act = 1;
			} else {
				# is this a usable screensaver?
				my ($sptr, $dptr) = &getSavers($saverType,$client);
				my @savers = @{$sptr};
				if ( defined($savers[$saverNumber]) ) {
					$act = 1;
					my $precheck = $preChangeFunctions{$savers[$saverNumber]};
					if ( defined($precheck) ) {
						$log->msg('funcs',60,"executing pre-usage check function for code $codename, saver pos $saverNumber, $savers[$saverNumber]");
						my %rhash = &$precheck( { client => $client, manual => 1 } );
						$act = $rhash{allow};
					}
				} else {
					$log->msg("player $id cannot use screensaver number $codename (pos $saverNumber)\n");
				}
			}
			# don't act if it hasn't been enough time
			if ( defined($lastChanged{$id}) && ((Time::HiRes::time() - $lastChanged{$id}) <= $minSwitchInterval) ) {
				$act = 0;
			}
			if ( $act == 1 ) {
				if ( (!defined($jumpButton{$id})) || ($codename ne $jumpButton{$id}) ) {
					$jumpButton{$id} = $codename;
					# set jumpToSaver timer
					$log->msg('timers',25,"$id: killing jumpToSaver() timers");
					Slim::Utils::Timers::killTimers($client, \&jumpToSaver);
					$log->msg('buttons',30,"button $codename pressed, will jump to specified saver in $Slim::Hardware::IR::IRSINGLETIME seconds");
					Slim::Utils::Timers::setTimer($client,Time::HiRes::time()+($Slim::Hardware::IR::IRSINGLETIME),\&jumpToSaver,
									$idstring,$saverNumber,$id,$saverType,$codename);
				}
			}
			# stop, don't pass this through!
			# (passing this along since it's not a held button will interrupt the screensaver)
			return;
		}
	} elsif ( defined($jumpButton{$id}) && ($jumpButton{$id} eq $codename) ) {
		$log->msg('buttons',50,"this code has not been checked yet, doing nothing");
		return;
	}
	$log->msg('buttons',70,"not interested in this event, calling original IR command");
        return &$originalIRCommand(@args);
}

sub jumpToSaver {
	my ($client,$idstring,$saverNumber,$id,$saverType,$codename) = @_;
	$jumpButton{$id} = undef;
	$lastIRcode{$id} = undef;
	$log->msg("jump to $saverType saver for button $codename\n");
	$manualOverride{$idstring} = $saverNumber;
	$lastChanged{$id} = Time::HiRes::time();
	&changeSaver($client,$saverType);
}

sub updatePref {
	my $id = shift;
	my $client = shift;
	#$log->msg("updatePref called for $id\n");
	if ( ($xPos{$id} == 1) && ($xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL|) ) {
		$log->msg('prefs',50,"buttonControl set to $yPos{$id}\n");
		Slim::Utils::Prefs::set("$screensaverName-button-control",$yPos{$id});
		return;
	}
	# get saverType
	my $saverType = &getMenuSaverType($id);
	my ($sptr, $dptr) = &getSavers($saverType,$client);
	my @savers = @{$sptr};
	my @durations = @{$dptr};
	# get saver, duration, saverMenuNpos
	my $duration = &getMenuValue($id,2);
	my $saverString = &getMenuValue($id,3);
	if ( $saverString eq '' ) {
		# have not arrowed that far yet, pull from settings
		my ($sptr, $dptr) = &getSavers($saverType,$client);
		my @savers = @{$sptr};
		my @durations = @{$dptr};
		# get plugin name from @savers
		my $name = $savers[$menuSaverNpos{$id}];
		# find l10n name from @saverStringNames/@saverNames
		$saverString = &getSaverString($name);
	}
	my $saver = &getSaverName($saverString);
	$savers[$menuSaverNpos{$id}] = $saver;
	$durations[$menuSaverNpos{$id}] = $duration;
	# set pref
	my $prefname = "$screensaverName-$saverType-savers";
	my $prefval = &makePrefString(\@savers,\@durations,$saverType);
	$log->msg("set $saverType savers to $prefval\n");
	Slim::Utils::Prefs::set($prefname,$prefval);
}

sub makePrefString {
	my ($sptr, $dptr, $saverType) = @_;
	my @savers = @{$sptr};
	my @durations = @{$dptr};
	my @stringArray = ();
	for (my $i = 0; $i < 9; ++$i) {
		my $d = $durations[$i];
		if ( (!defined($d)) || ($d eq '') ) { $d = 0; }
		my $s = $savers[$i];
		if ( (!defined($s)) || ($s eq '') ) { $s = $defaultSavers{$saverType}; }
		push @stringArray, CGI::escape($d).':'.CGI::escape($s);
	}
	return join("\t",@stringArray);
}

sub initConfig {
	#$log->msg("initConfig running\n");
	# get pref
	my $buttonControl = Slim::Utils::Prefs::get("$screensaverName-button-control");
	if ( (!defined($buttonControl)) || ($buttonControl eq '') ) {
		$log->msg('prefs',25,"buttonControl set to 1\n");
		Slim::Utils::Prefs::set("$screensaverName-button-control",1);
	}
	my $pref = Slim::Utils::Prefs::get("$screensaverName-offsaver-savers");
	# if not existing, set defaults
	if ( defined($pref) && ($pref ne '') ) {
		# make sure we have 9 items
		&padConfig;
		# all set
		return;
	}
	$log->msg("initConfig setting defaults\n");

	# defaults
	
	# off:
	# auto: datetime 
	# manual: RSS, snow
	Slim::Utils::Prefs::set("$screensaverName-offsaver-savers","60:SCREENSAVER.datetime\t0:SCREENSAVER.rssnews\t0:SCREENSAVER.snow");
	
	# idle:
	# auto: datetime 30s, RSS 60s 
	# manual: snow
	Slim::Utils::Prefs::set("$screensaverName-idlesaver-savers","30:SCREENSAVER.datetime\t60:SCREENSAVER.rssnews\t0:SCREENSAVER.snow");
	
	# playing:
	# auto: playlist 30s, datetime 15s
	# manual: analog VU, digitatl VU, spectrum analyzer
	Slim::Utils::Prefs::set("$screensaverName-screensaver-savers","30:playlist\t15:SCREENSAVER.datetime\t:0:SCREENSAVER.visualizer_analog_vumeter\t0:SCREENSAVER.visualizer_digital_vumeter\t0:SCREENSAVER.visualizer_spectrum");
	
	# make sure each pref string has 9 entries	
	&padConfig;
	
	# Common screensaver names/l10n string names
	# name "SCREENSAVER.datetime" string name "PLUGIN_SCREENSAVER_DATETIME"
	# name "SCREENSAVER.plugin_FuzzyTime" string name "PLUGIN_FUZZYTIME"
	# name "SCREENSAVER.rssnews" string name "PLUGIN_RSSNEWS_SCREENSAVER"
	# name "SCREENSAVER.snow" string name "PLUGIN_SCREENSAVER_SNOW"
	# name "SCREENSAVER.superdatetime" string name "PLUGIN_SCREENSAVER_SUPERDATETIME"
	# name "SCREENSAVER.visualizer_analog_vumeter" string name "VISUALIZER_ANALOG_VUMETER"
	# name "SCREENSAVER.visualizer_digital_vumeter" string name "VISUALIZER_DIGITAL_VUMETER"
	# name "SCREENSAVER.visualizer_spectrum" string name "VISUALIZER_SPECTRUM_ANALYZER"
	# name "SCREENSAVER.xdatetime" string name "PLUGIN_SCREENSAVER_XDATETIME"
	# name "playlist" string name "NOW_PLAYING"
	#
	# and some that aren't really usable:
	# name "nosaver" string name "SCREENSAVER_NONE"
	# name "screensaver" string name "SCREENSAVER_JUMP_BACK_NAME"
}

sub padConfig() {
	foreach my $type ('offsaver', 'idlesaver', 'screensaver') {
		my $pref = Slim::Utils::Prefs::get("$screensaverName-${type}-savers");
		my @items = split(/\t/, $pref);
		# remove extra items
		if ( scalar(@items) > 9 ) {
			&Plugin_msg("removing extra $type screensaver entries --  ".scalar(@items)." is too many\n");
		}
		for (my $i = scalar(@items); $i > 9; --$i ) {
			pop @items;
		}
		# remove invalid items
		my @newitems = ();
		for (my $i = 0; $i < scalar(@items); ++$i ) {
			if ( $items[$i] =~ m|^[0-9]{1,}\:.{1,}$| ) {
				push @newitems, $items[$i];
			} else {
				&Plugin_msg("removing invalid $type screensaver entry \"".$items[$i]."\"\n");
			}
		}
		# add more default savers to get to a count of 9
		for (my $i = scalar(@newitems); $i < 9; ++$i ) {
			my $saver = $defaultSavers{$type};
			push @newitems, "0:".CGI::escape($saver);
		}
		my $newpref = join("\t",@newitems);
		if ( $newpref ne $pref ) {
			&Plugin_msg("${type}: change pref from \"$pref\" to \"$newpref\"\n");
			Slim::Utils::Prefs::set("$screensaverName-${type}-savers",$newpref);
		}
	}
}

sub getMenuValue {
	my ($id,$xpos) = @_;
	my $hashString = $xHash{$id};
	my @escapedValues = split(/\t/, $hashString);
	my $rc = CGI::unescape($escapedValues[$xpos]);
	return $rc;
}

sub getPlayerLevel {
	my $client = shift;
	if (! defined($client) ) { return -1; }
	if ( $client->isa( "Slim::Player::Squeezebox2") ) { return 3; }
	if ( $client->isa( "Slim::Player::Transporter") ) { return 3; }
	if ( $client->isa( "Slim::Player::SLIMP3") ) { return 0; }
	if ( $client->isa( "Slim::Player::SoftSqueeze") ) { return 3; }
	# BUG: how to distinguish a SB1 from an SBG?
	if ( $client->isa( "Slim::Player::Squeezebox") ) { return 1; }
	# default: unknown
	return -1;
}

sub getSavers($$) {
	my $saverType = shift;
	my $id;
	my $playerLevel;
	my $client = shift;
	if ( defined($client) ) {
		$id = $client->id();
		$playerLevel = &getPlayerLevel($client);
	} else {
		# no client, do not do capability filtering
		$id = "ignored";
		$playerLevel = 99;	# impossibly high value so we'll pass all saverMinimums tests
	}
	#$log->msg("getting $saverType savers\n");
	# get pref
	my $pref = Slim::Utils::Prefs::get("$screensaverName-$saverType-savers");
	# start with empty arrays
	my @savers = ();
	my @durations = ();
	# loop through and add to arrays where duration > 0
	my @escapedValues = split(/\t/, $pref);
	my $found = 0;
	my $activefound = 0;
	for (my $i = 0; $i < scalar(@escapedValues); ++$i) {
		my ($ed,$es) = split(/\:/,$escapedValues[$i]);
		my $d = CGI::unescape($ed);
		my $s = CGI::unescape($es);
		my $sString = &getSaverString($s);
		if ( ($d ne '') && (defined($sString)) ) {
			# saver looks good, but can we use it?
#$log->msg("player $id level $playerLevel, considering $s\n");
			if ( defined($saverMinimums{$s}) && ($saverMinimums{$s} > $playerLevel) ) {
				$log->msg("player $id cannot use screensaver $s/$sString\n");
				$s = undef;
				--$found;
			}
			if ( ($d > 0) && ($d < $minActiveDuration) ) { $d = $minActiveDuration; }
			push @durations, $d;
			push @savers, $s;
			++$found;
			if ( $d > 0 ) { ++$activefound; }
		}
	}
	# once done, if no active savers found, add something at 60 seconds
	if ( $activefound == 0 ) {
		$log->msg('misc',60,"no active savers for \"$saverType\" found, adding default as 10th saver");
		push @durations, 60;
		push @savers, $defaultSavers{$saverType};
	}
	return ( \@savers, \@durations );
}

sub setScreensaverMode() {
	my $client = shift;
	my $id = $client->id();
	#$log->msg("setScreensaverMode called\n");
	my $clientMode = Slim::Buttons::Common::mode($client);
	$clientMode =~ s/\..*$//;
	$expectedMode{$id} = $clientMode;
	my $saverType = &whichSaverType($client);
	if (! defined($saverPos{$id}) ) {
		$saverPos{$id} = -1;
		$saverPos{$id} = &nextActiveSaver($id,$saverType,$client);
	}
	# sometimes called for wrong type?!?
	my $currentSaver = $client->prefGet($saverType);
	if ( $currentSaver ne $screensaverName ) {
		$log->msg("$id: called for wrong saver? ($currentSaver ne $screensaverName) exiting\n");
		$log->msg('misc',60,"$id: called for wrong saver? ($currentSaver ne $screensaverName) exiting\n");
		if ( defined($multiMode{$id}) ) {
			&noteAllPlayers($multiMode{$id});
			&resetPlayerPrefs(undef,$id);
			$log->msg('timers',25,"$id: setScreensaverMode() killing lookForPlaylistChanges() timers");
			Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
			$saverPos{$id} = undef;
			$multiMode{$id} = undef;
		}
		return;
	}
	if (! defined($multiMode{$id}) ) {
		$multiMode{$id} = $saverType;
	} elsif ( $multiMode{$id} ne $saverType ) {
		# called as a diff screensaver type!
		$log->msg("killing SaverSwitcher timer for $id since now called as a $saverType\n");
		$log->msg('misc',60,"killing SaverSwitcher timer for $id since now called as a $saverType\n");
		$log->msg('timers',25,"$id: setScreensaverMode() killing changeSaver() timers");
		Slim::Utils::Timers::killTimers($client, \&changeSaver);
		&noteAllPlayers($multiMode{$id});
		&resetPlayerPrefs(undef,$id);
		$log->msg('timers',25,"$id: setScreensaverMode() killing lookForPlaylistChanges() timers");
		Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
		$multiMode{$id} = $saverType;
		# need to start at first saver
		$saverPos{$id} = -1;
		$saverPos{$id} = &nextActiveSaver($id,$saverType,$client);
	}
	my $notePlayer = 0;
	my $idstring = "$id\t$saverType";
	# set saver pref, set timer for next mode
	&delegateSaver($client,$id,$saverType,$idstring);
	# the following makes the display show "Loading...", 
	# but a blank screen seems nicer
	#$client->lines(\&screensaverLines);
}

sub delegateSaver {
	my ($client,$id,$saverType,$idstring,$specificPos) = @_;
	my $idstring = "$id\t$saverType";
	my ($sptr, $dptr) = &getSavers($saverType,$client);
	my @savers = @{$sptr};
	my @durations = @{$dptr};
	if ( 
		(!defined($saverPos{$id})) || 
		($saverPos{$id} >= scalar(@savers)) || 
		($saverPos{$id} < 0)
	   ) {
		$log->msg("returning to first screensaver\n");
		$saverPos{$id} = -1;
		$saverPos{$id} = &nextActiveSaver($id,$saverType,$client);
	}
	# use manual overrride if set
	if ( defined($manualOverride{$idstring}) ) {
		$saverPos{$id} = $manualOverride{$idstring};
	}
	# unless passed a specific saver # (as for new song)
	if ( defined($specificPos) ) {
		$saverPos{$id} = $specificPos;
	}
	my $delay = $durations[$saverPos{$id}];
	if ( defined($durationToUse{"$id-$saverPos{$id}"}) ) {
		# use the override duration
		$delay = $durationToUse{"$id-$saverPos{$id}"};
		# and forget the override
		$durationToUse{"$id-$saverPos{$id}"} = undef;
	}
	my $saver = $savers[$saverPos{$id}];
	# make sure we know $client uses SaverSwitcher for $saverType
	#if ( (!defined($playerSaver{$idstring})) || ($playerSaver{$idstring} eq '') ) {
#$log->msg("noting that $idstring is controlled by SaverSwitcher\n");
	$playerSaver{$idstring} = $saver;
	&noteAllPlayers($saverType); 
	if ( (defined($manualOverride{$idstring}) && defined($specificPos)) ||
	     (!defined($manualOverride{$idstring})) ) {
		$log->msg("$id: setting $saverType saver to $saver, changing in $delay seconds\n");
	} else {
		$log->msg("$id: setting $saverType saver to $saver as requested\n");
	}
	$client->param('modeUpdateInterval', 1); # seconds
	$client->prefSet($saverType,$saver);
	$multiMode{$id} = $saverType;
	if ( $saverType eq 'screensaver' ) {
		# start looking for appearance of new songs
		&startLookingForPlaylistChanges($client,$id);
	}
	$log->msg('timers',25,"$id: delegateSaver() killing changeSaver() timers");
	Slim::Utils::Timers::killTimers($client, \&changeSaver);
	if ( (defined($manualOverride{$idstring}) && defined($specificPos)) ||
	     (!defined($manualOverride{$idstring})) ) {
		$log->msg('timers',25,"$id: delegateSaver() resetting changeSaver() timers");
		Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $delay, \&changeSaver, $saverType);
	}
}

sub nextActiveSaver($$$) {
	my ($id, $saverType, $client) = @_;
	my ($sptr, $dptr) = &getSavers($saverType,$client);
	my @savers = @{$sptr};
	my @durations = @{$dptr};
	my $startPos = $saverPos{$id};
	if (! defined($startPos) ) { $startPos = -1; }
	for (my $i = 1; $i <= scalar(@savers); ++ $i ) {
		my $pos = ( $startPos + $i ) % scalar(@savers);
		$durationToUse{"$id-$pos"} = undef;
		$log->msg('misc',70,"considering saver $pos for $saverType");
		if ( ($pos != 9) && ($pos >= 0) && ($durations[$pos] > 0) && defined($savers[$pos]) ) {
			my $ok = 1;
			my $precheck = $preChangeFunctions{$savers[$pos]};
			if ( defined($precheck) ) {
				$ok = 0;
				$log->msg('funcs',60,"executing pre-usage check function for saver pos $pos, $savers[$pos]");
				my %rhash = &$precheck( { client => $client, manual => 0, minDuration => $minActiveDuration, maxDuration => $durations[$pos] } );
				if ( $rhash{allow} == 1 ) {
					$ok = 1;
					my $duration = $rhash{duration};
					if ( defined($duration) && ($duration < $durations[$pos]) ) {
						$durationToUse{"$id-$pos"} = ($duration < $minActiveDuration) ? $minActiveDuration : $duration;
					}
				}
			}
			if ( $ok == 1 ) { return $pos; }
		}
	}
	# pos 9 should always be OK ("#10", from padding)
	return 9;
	# none found -- stay where you are
	return $startPos;
}

sub changeSaver {
	my ($client,$saverType,$specificPos) = @_;
	$::d_plugins && defined($specificPos) && &Plugin_msg("asked to use pos $specificPos\n");
	my $id = $client->id();
	my $idstring = "$id\t$saverType";
	#$log->msg("changeSaver called\n");
	my $clientMode = Slim::Buttons::Common::mode($client);
	$clientMode =~ s/\..*$//;
	# bail out if user changed the screensaver
	my $currentSaver = $client->prefGet($saverType);
	if ( defined($playerSaver{$idstring}) && ($playerSaver{$idstring} ne $currentSaver) && ($currentSaver ne $screensaverName) ) {
		$log->msg("it appears the user chose $currentSaver for $saverType\n");
		# set $playerSaver{$idstring} to '' to indicate that SaverSwitcher doesn't control this client+saverType
		$log->msg("noting that $idstring is NOT controlled by SaverSwitcher\n");
		$playerSaver{$idstring} = '';
		# reset our persistent list of SaverSwitcher playes for $saverType
		&noteAllPlayers($saverType);
		$manualOverride{$idstring} = undef;
		# bail out without acting
		$multiMode{$id} = undef;
		return;
	}
	# if not in screensaver mode, reset & return
	# BUG -- I've had a fair bit of trouble reliably ascertaining
	# if the client is still in a screensaver
	my $currentSaverWouldBe = &whichSaverType($client);
	$log->msg('misc',90,"expectedMode{id} \"$expectedMode{$id}\" clientMode \"$clientMode\" playerSaver{idstring} \"$playerSaver{$idstring}\" currentSaverWouldBe \"$currentSaverWouldBe\" saverType \"$saverType\"\n");
	if ( 
	     (!defined($expectedMode{$id})) || 
	     ( 
	       ($saverType eq 'idlesaver') &&
	       ($clientMode ne $expectedMode{$id}) && 
               ($clientMode ne 'SCREENSAVER') && 
               ($clientMode ne $playerSaver{$idstring})
	     ) ||
	       (($saverType eq 'offsaver') && ($currentSaverWouldBe ne $saverType)) ||
	       (($saverType eq 'screensaver') && ($currentSaverWouldBe ne $saverType))
           ) {
		$log->msg("$id: current mode $clientMode not $expectedMode{$id} would be $currentSaverWouldBe should be $saverType -- out of screensaver mode?\n");
		$log->msg('misc',60,"$id: current mode $clientMode not $expectedMode{$id} would be $currentSaverWouldBe should be $saverType -- out of screensaver mode?\n");
		#&noteAllPlayers($saverType);
		&resetPlayerPrefs(undef,$id);
		$log->msg('timers',25,"$id: changeSaver() killing lookForPlaylistChanges() timers");
		Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
		$manualOverride{$idstring} = undef;
		$multiMode{$id} = undef;
		$saverPos{$id} = undef;
		return;
	}
	# don't change if we're about to change b/c of a new song
	if ( ($saverType eq 'screensaver') && (&getSongId($client) ne $latestSong{$id}) ) {
		# new song; wait and let the automatic change logic act
		$log->msg('misc',60,"$id: new song; wait and let the automatic change logic act");
		return;
	}
	# move to the next saver unless told a specific one to use
	if ( (!defined($manualOverride{$id})) && (!defined($specificPos)) ) {
		$saverPos{$id} = &nextActiveSaver($id,$saverType,$client);
	}
	# set saver pref, set timer for next mode
	$log->msg('misc',60,"$id: calling delegateSaver");
	&delegateSaver($client,$id,$saverType,$idstring,$specificPos);
}

# verify that this notices updated metadata in streams
sub getSongId($) {
	my $client = shift;
	my $song = Slim::Player::Playlist::song($client);
	# return several pieces of info, so it's more likely to recognize
	# changes if playing several recordings of the same song in a row
	# The following is good for local recordings
	my $songinfo = join("\t",$song->name(),$song->duration(),$song->path()); 
	# but we need this to get the title from stream metadata
	# (only add this for streams; for local recordings, the following changes
	#  before the music really starts playing[!])
	if ( $song->path() =~ m|^[a-z0-9]{2,}:|i ) {
		my $url = Slim::Player::Playlist::url( $client, Slim::Player::Source::streamingSongIndex($client) );
		$songinfo .= "\t".Slim::Music::Info::getCurrentTitle(undef, $url)."\tURL";
	} else {
		$songinfo .= "\tFILE";
	}
	$log->msg("currently playing \"$songinfo\"\n");
	return $songinfo;
}

sub getFirstNowPlayingSaver {
	my $client = shift;
	my ($sptr, $dptr) = &getSavers('screensaver',$client);
	my @savers = @{$sptr};
	for (my $i = 0; $i < scalar(@savers); ++$i) {
		# name "playlist" string name "NOW_PLAYING"
		if ($savers[$i] eq 'playlist') { return $i; }
	}
	# not found
	return -1;
}

sub secondsLeft($) {
	my $client = shift;
	my $id = $client->id();
	# from Slim::Player::Source
	my $rate        = $client->rate();
	if (Slim::Player::Source::playingSongIndex($client) == Slim::Player::Source::streamingSongIndex($client)) {
		my $outputBufferFullness = $client->outputBufferFullness();
		if (defined($outputBufferFullness)) {
			# Assume 44.1KHz output sample rate. This will be slightly
			# off for anything that's 48Khz, but it's a guesstimate anyway.
			return (($outputBufferFullness / (44100 * 8)) * $rate);
		}
	}
	return 0;
}

sub startLookingForPlaylistChanges($$) {
	my ($client,$id) = @_;
	$log->msg('timers',25,"$id: startLookingForPlaylistChanges() killing lookForPlaylistChanges() timers");
	Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
	if ( ! defined($latestSong{$id}) ) {
		# note the current playlist text
		my $song = &getSongId($client);
		#$log->msg("$id is currently playing \"$song\"\n");
		$latestSong{$id} = $song;
	}
	if ( $latestSong{$id} !~ m/\tFILE$/ ) {
		# set a timer for lookForPlaylistChanges in case playlist.newsong doesn't fire
		Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $lookInterval, \&lookForPlaylistChanges);
	}
}

sub lookForPlaylistChanges {
	my $client = shift;
	my $id = $client->id();
	my $idstring = "$id\tscreensaver";
	my $delay = 0;
	# verify still in SaverSwitcher controlled 'screensaver' mode
	if ( (&whichSaverType($client) ne 'screensaver') || (!defined($playerSaver{$idstring})) || ($playerSaver{$idstring} eq '') ) {
		# don't worry about it
		return;
	}
	# note the current playlist text
	my $song = &getSongId($client);
	#$log->msg("$id is currently playing \"$song\"\n");
	# if different than last seen text
	if ( $song ne $latestSong{$id} ) {
		# how much time left in the buffer? (for streams only)
		if ( $song =~ m|URL$| ) {
			$delay = &secondsLeft($client);
		}
		# add a wee bit of time to account for song startup, recognition of transition, etc.
		# (we should have noticed the info somewhere between 0 and $lookInterval seconds after the song started)
		$delay += 1.5;
		# note new text
		# go back to saver #0 ?
		my ($sptr, $dptr) = &getSavers('screensaver',$client);
		my @savers = @{$sptr};
		my @durations = @{$dptr};
		if ( defined($durations[0]) && defined($savers[0]) && ($durations[0] > 0) && ($savers[0] ne '') ) {
			$log->msg("$id is playing a new song \"$song\", returning to saver 0 in $delay seconds\n");
			$latestSong{$id} = $song;
			# kill any changeSaver timers
			$log->msg('timers',25,"$id: new song, lookForPlaylistChanges killing changeSaver() timers");
			Slim::Utils::Timers::killTimers($client, \&changeSaver);
			Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $delay, \&changeSaver, 'screensaver',0);
			# changeSaver() will call delegateSaver() which will schedule the next lookForPlaylistChanges() call
		} else {
			$latestSong{$id} = $song;
			# set a timer for this to run again in a few seconds
			# in case screensaver #0 (1/9) is activated before the next song change
			$log->msg("$id is playing a new song \"$song\", but saver 0 is not active\n");
			$log->msg('timers',25,"$id: lookForPlaylistChanges killing lookForPlaylistChanges() timers");
			Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
			Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $lookInterval, \&lookForPlaylistChanges);
		}
	} else {
		# set a timer for this to run again in a few seconds (ideally, buffer size + 1?)
		$log->msg('timers',25,"$id: lookForPlaylistChanges resetting lookForPlaylistChanges() timers");
		Slim::Utils::Timers::killTimers($client, \&lookForPlaylistChanges);
		Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $lookInterval, \&lookForPlaylistChanges);
	}
}

# note all players using SaverSwitcher for a given
# screensaver type so we can reset their prefs at server startup
sub noteAllPlayers {
	my $saverType = shift;
	#$log->msg("noteAllPlayers($saverType) called\n");
	my $playerListString = Slim::Utils::Prefs::get("$screensaverName-$saverType-players");
	foreach my $idstring (keys %playerSaver) {
		if ($idstring =~ m/^(.*)\t${saverType}/) {
			my $id = $1;
			my $add = 1;
			if ( $playerSaver{$idstring} ne '' ) {
				# SaverSwitcher has been handling this
				my $client = Slim::Player::Client::getClient($id);
				if ( defined($client) ) {
					my $currentSaver = $client->prefGet($saverType);
					if ( ($currentSaver ne $screensaverName) && ($currentSaver ne $playerSaver{$idstring}) ) {
						$add = 0;
					}
				}
			} else {
				# SaverSwitcher acknowledges this screensaver is no longer used
				$add = 0;
			}	
			if ( $add == 1 ) {
				if ( $playerListString !~ m/${id}\t/ ) {
					$log->msg("adding $id to list of players using SaverSwitcher for $saverType\n");
					$playerListString .= "${id}\t";
				}
			} else {
				# if in string, remove it
				$log->msg("removing $id from list of players using SaverSwitcher for $saverType\n");
				$playerListString =~ s/${id}\t//;
			}
		}
	}
	# reset our pref string
	Slim::Utils::Prefs::set("$screensaverName-$saverType-players",$playerListString);
}

sub newSong {
	my $request = shift;
	$log->msg("newSong callback running\n");
	my $client = $request->client();
	if ( defined($client) ) {
		my $id = $client->id();
		if ( defined($id) ) {
			&lookForPlaylistChanges($client);
		}
	}
}

sub newPlayerReset {
	my $request = shift;
	$log->msg("newPlayerReset callback running\n");
	my $client = $request->client();
	if ( defined($client) ) {
		my $id = $client->id();
		if ( defined($id) ) {
			&resetPlayerPrefs(undef,$id);
		}
	}
}

# reset all player prefs
sub resetPlayerPrefs {
	my ($delay,$specificId) = @_;
	#$log->msg("resetPlayerPrefs called\n");
	my $unfoundPlayerCount = 0;
	my $unfoundPlayerCount = 0;
	my %unfoundPlayers;
	foreach my $saverType ( 'offsaver', 'screensaver', 'idlesaver' ) {
		my $playerListString = Slim::Utils::Prefs::get("$screensaverName-$saverType-players");
		my @ids = split(/\t/,$playerListString);
		foreach my $id ( @ids ) {
			if ( (!defined($specificId)) || ($specificId eq $id) ) {
				my $idstring = "$id\t$saverType";
				my $client = Slim::Player::Client::getClient($id);
				if (defined($client)) {
					#$log->msg("restoring $saverType player $id\n");
					my $currentSaver = $client->prefGet($saverType);
					if ($currentSaver ne $screensaverName) {
						$log->msg("Resetting $saverType for $id to SaverSwitcher\n");
						$client->prefSet($saverType,$screensaverName);
						$multiMode{$id} = undef;
						$manualOverride{$idstring} = undef;
						# remove from this list in case we keep running this sub for missing players
						$playerListString =~ s/${id}\t//;
						Slim::Utils::Prefs::set("$screensaverName-$saverType-players",$playerListString);
					}
				} else {
					if (!defined($unfoundPlayers{$id}) ) {
						#$log->msg("player $id is not connected\n");
						++$unfoundPlayerCount;
						$unfoundPlayers{$id} = 1;
					}
				}
			}
		}
	}
	if ( $unfoundPlayerCount > 0 ) {
		if ( defined($delay) ) {
			$delay *= 1.5;
			if ( $delay > 60 ) { $delay = 60; }
			$log->msg("$unfoundPlayerCount players not found, will try again in $delay seconds\n");
			$log->msg('timers',25,"resetting resetPlayerPrefs() timers");
			Slim::Utils::Timers::killTimers($delay, \&resetPlayerPrefs);
			Slim::Utils::Timers::setTimer($delay, Time::HiRes::time() + $delay, \&resetPlayerPrefs);
		}
	} else {
		if (! defined($specificId) ) {
			$log->msg("all player screensaver preferences have been restored\n");
		}
	}
}

sub whichSaverType {
	my $client = shift;
	my $id = $client->id();
	if ($client->power() == 0) { 
		$log->msg('misc',90,"$id is powered off\n");
		return 'offsaver'; 
	}
	# from Screensaver.pm:
	my $mode = Slim::Player::Source::playmode($client);
	if ( $mode eq 'play' ) {
		$log->msg('misc',90,"$id is playing music\n");
		return 'screensaver';
	}
	$log->msg('misc',90,"$id is on, but not playing music\n");
	return 'idlesaver';
}

sub leaveScreensaverMode() {
	my $client = shift;
	my $id = $client->id();
	#$log->msg("leaveScreensaverMode called\n");
}

our %screensaverFunctions = (
);

sub getScreensaverFunctions() {
	return \%screensaverFunctions;
}

sub Plugin_msg {
	my $message = shift;
	msg("SaverSwitcher: $message");
}

sub screenSaver {
	my $client = shift;
	Slim::Buttons::Common::addSaver($screensaverName,
		getScreensaverFunctions(),
		\&setScreensaverMode,
		\&leaveScreensaverMode,
		'PLUGIN_SAVERSWITCHER_SAVER');
}

# this should never run
sub screensaverLines {
	my $client = shift;
	my $line1 = string('PLUGIN_SAVERSWITCHER');
	my $line2 = string('LOADING');
	return { 'center' => [$line1, $line2] };
}

sub getConfig {
	my $client = shift;
	my $id = $client->id();
}	

# player UI code

sub addMenu {
	return "PLUGINS";
}

sub setMode {
        my $client = shift;
	$client->lines(\&displayLines);
}

# grid:
# x = 0  -> off vs idle vs while playing
# x = 1  -> savers 1 to N 
# x = 2  -> choose a duration for saver N for position x2 for saver mode x1
# 	    if duration == 0, then only (y-1) savers
# 	    if duration > 0, allow another saver
# x = 3  -> choose a saver for position x2 for mode x1
# 	    MUST default to current choice

# tracking:
# each player has an array of string names, e.g.
#  [ 'SETUP_OFFSAVER', 'SAVER_0', 'PLUGIN_SCREENSAVER_DATETIME', '30' ]
# and a single variable for the X position
#
# moving left, use the pos corresponding to the string name if possible
# moving right,  check current settings & default to proper position, e.g
#   moving from 1 to 2, see what the current saver is and start at that Y2 position
#   moving from 2 to 3, see what the current delay is and start with that


our %menuFunctions = (
        # normal up/right/left/down navigation:
        'left' => sub  {
                my $client = shift;
		my $button = shift;
		$log->msg('menu',80, "$button pressed");
		if ( $button ne 'left' ) { return; }
		my $id = $client->id();
		$log->msg('menu',70, " left, hash at start is \"$xHash{$id}\"");
		if ( &moveXRight($id,-1,$client) < 0 ) {
			$log->msg('menu',80, "leaving SaverSwitcher mode");
                	Slim::Buttons::Common::popModeRight($client);
		} else {
			$client->lines(\&displayLines);
			$client->update();
		}
		$log->msg('menu',70, " left, hash at end   is \"$xHash{$id}\"");
        },
        'right' => sub  {
                my $client = shift;
		my $button = shift;
		$log->msg('menu',80, "$button pressed");
		if ( $button ne 'right' ) { return; }
		my $id = $client->id();
		$log->msg('menu',70, "right, hash at start is \"$xHash{$id}\"");
		if ( &moveXRight($id,1,$client) < 0 ) {
			$log->msg('menu',80, "cannot move right");
                	$client->bumpRight($client);
		} else {
			$client->lines(\&displayLines);
			$client->update();
		}
		$log->msg('menu',70, "right, hash at end   is \"$xHash{$id}\"");
        },
        'up' => sub  {
                my $client = shift;
		my $button = shift;
		$log->msg('menu',80, "$button pressed");
		if ( ($button ne 'up') ) { return; }
		my $id = $client->id();
		$log->msg('menu',70, "   up, hash at start is \"$xHash{$id}\"");
		&moveYDown($id,-1,$client);
		$client->lines(\&displayLines);
		$client->update();
		$log->msg('menu',70, "   up, hash at end   is \"$xHash{$id}\"");
        },
        'down' => sub  {
                my $client = shift;
		my $button = shift;
		$log->msg('menu',80, "$button pressed");
		if ( ($button ne 'down') ) { return; }
		my $id = $client->id();
		$log->msg('menu',70, " down, hash at start is \"$xHash{$id}\"");
		&moveYDown($id,1,$client);
		$client->lines(\&displayLines);
		$client->update();
		$log->msg('menu',70, " down, hash at end   is \"$xHash{$id}\"");
        },
);

sub moveXRight {
	my ($id,$amount,$client) = @_;
	$log->msg('funcs',50, "moveXRight called");
	$log->msg('menu',80, "$id moveXRight called");
	&initMenuPos($id);
	if ( ($amount > 0) && ($xPos{$id} == 1) && ($xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL|) ) {
		# can't move right from here
		$log->msg('funcs',50, "moveXRight done");
		return -1;
	}
	#if ( ($xPos{$id} == 2) && ($yPos{$id} == 0) && ($amount > 0) ) { 
	#	# no sense moving right if this is disabled?
	#	$log->msg('funcs',50, "moveXRight done");
	#	return -1;
	#}
	$log->msg('funcs',80, "moveXRight add $amount to $xPos{$id}");
	$xPos{$id} += $amount;
	if ( $xPos{$id} < 0 ) { 
		$xPos{$id} = 0; 
		$log->msg('menu',80, "$id cannot move left");
		$log->msg('funcs',50, "moveXRight done");
		return -1;
	}
	if ( $xPos{$id} > 3) { 
		$xPos{$id} = 3; 
		$log->msg('menu',80, "$id cannot move right");
		$log->msg('funcs',50, "moveXRight done");
		return -1;
	}
	if ( $amount > 0 ) {
		$log->msg('menu',50, "moving right, look for good position to start at");
		# Y pref we should default to?
		
		# the following would use the remembered N/9 pos			
		#if ( ($xPos{$id} == 1) && (defined($menuSaverNpos{$id})) ) {
		#	$log->msg('menu',60, "$id in pos 1, start at yPos = menuSaverNpos $menuSaverNpos{$id}");
		#	$yPos{$id} = $menuSaverNpos{$id};
		#} elsif ( $xPos{$id} < 2 ) {
		# but it seems cleaner to just always right-arrow from saverType to the 1/9 item
		if ( ($xPos{$id} == 1) && ($yPos{$id} == 3) ) {
			$log->msg('menu',60, "$id at button control");
			$yPos{$id} = Slim::Utils::Prefs::get("$screensaverName-button-control");
		} elsif ( $xPos{$id} < 2 ) {
			$log->msg('menu',60, "$id in pos 0 or 1, start at yPos 0");
			$yPos{$id} = 0;
		} else {
			my $saverType = &getMenuSaverType($id);
			# get pos in $menuItems[0]
			# get pref name
			my ($sptr, $dptr) = &getSavers($saverType,undef);
			my @savers = @{$sptr};
			my @durations = @{$dptr};
			if ( $menuSaverNpos{$id} < scalar(@savers) ) {
				if ( $xPos{$id} == 3 ) {
					$log->msg('menu',60, "$id looking up name for screensaver at pos $menuSaverNpos{$id}");
					# get plugin name from @savers
					my $name = $savers[$menuSaverNpos{$id}];
					# find l10n name from @saverStringNames/@saverNames
					my $string = &getSaverString($name);
					if (!defined($string)) { $string = &getSaverString($savers[0]); }
					#my $string = $savers[$menuSaverNpos{$id}];
					# find y pos from @{$menuItems[$xPos{$id}]};
					$yPos{$id} = &getSaverPos($string);
					$log->msg('menu',60, "$id start at yPos $yPos{$id} for saver $string");
				} else {
					# x3
					$yPos{$id} = $durations[$menuSaverNpos{$id}];
					$log->msg('menu',60, "$id start at duration value $yPos{$id}");
				}
			} else {
				$yPos{$id} = 0;
				$log->msg('menu',60, "$id start at fallback yPos $yPos{$id}");
			}
		}
		# update X name hash
		my $hashCount = scalar(split(/\t/,$xHash{$id}));
		if ( $hashCount > $xPos{$id} ) {
			# trim before adding
			$log->msg('menu',60, "removing last item from \"$xHash{$id}\"");
			$xHash{$id} =~ s/[^\t]*\t$//;
		}
		if ( ($xPos{$id} == 1) && ($xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL|) ) {
			$log->msg('menu',60, "adding button control $yPos{$id} to \"$xHash{$id}\"");
			$xHash{$id} .= CGI::escape($yPos{$id})."\t";
		} elsif ( $xPos{$id} == 2 ) {
			$log->msg('menu',60, "adding duration $yPos{$id} to \"$xHash{$id}\"");
			$xHash{$id} .= CGI::escape($yPos{$id})."\t";
		} else {
			my $token = ${$menuItems[$xPos{$id}]}[$yPos{$id}];
			$log->msg('menu',60, "adding token \"$token\" to \"$xHash{$id}\"");
			$xHash{$id} .= CGI::escape($token)."\t";
		}
	} else {
		# amount < 0, moving left
		# try to find where we were
		$yPos{$id} = 0;
		$log->msg('menu',50, "moving left, trying to find proper position in \"$xHash{$id}\"");
		my $hashCount = scalar(split(/\t/,$xHash{$id}));
		my $lastVal = &getMenuValue($id,$xPos{$id});
		# read X hash, reset Y value to match
		my $arrPtr = $menuItems[$xPos{$id}];
		if (defined($arrPtr)) {
			my @choices = @{$arrPtr};
			for (my $i = 0; $i < scalar(@choices); ++$i ) {
				$log->msg('menu',80, "at $i, compare $choices[$i] to $lastVal");
				if ( $choices[$i] eq $lastVal ) { 
					$log->msg('menu',70, "at $i, $choices[$i] matches $lastVal");
					$yPos{$id} = $i;
				}
			}
		}
		# special case for duration
		if ( $xPos{$id} == 2 ) {
			$log->msg('menu',70, "duration spot: use $lastVal");
			$yPos{$id} = $lastVal;
		}
		# if now at pos 1, clear last xHash so updatePrefs() gets correct name
		if ( ($xPos{$id} == 1) || ($xPos{$id} == 0) ) {
			$log->msg('menu',50, "back to xPos 1, clearing xHash duration/saver info");
			my @xArray = split(/\t/,$xHash{$id});
			while ( scalar(@xArray) > ($xPos{$id} + 1) ) {
				pop @xArray;
			}
			$xHash{$id} = join("\t",@xArray)."\t";
		}
	}
	if ( $xPos{$id} == 1 ) {
		$log->msg('menu',50, "setting menuSaverNpos{$id} to $yPos{$id}");
		$menuSaverNpos{$id} = $yPos{$id};
	}
	$log->msg('funcs',50, "moveXRight done");
	return $xPos{$id};
}

sub getMenuSaverType {
	my $id = shift;
	$log->msg('menu',80, "getMenuSaverType($id)");
	$xHash{$id} =~ m/^([^\t]{1,})\t/;
	my $escaped = $1;
	# get saverType l10n string from beginning of $xHash
	my $saverTypeString = CGI::unescape($escaped);
	my $saverType = &getSaverType($saverTypeString);
	return $saverType;
}

sub moveYDown {
	my ($id,$amount,$client) = @_;
	&initMenuPos($id);
	$log->msg('menu',80, "$id moveYDown($amount)");
	my $maxPos;
	my $update = 0;
	my @choices = ();
	if ( ($xPos{$id} == 1) && ($xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL|) ) {
		$log->msg('menu',80, "will update prefs");
		$update = 1; 
		@choices = @buttonChoices;
	}
	if ( $xPos{$id} > 1 ) { 
		$log->msg('menu',80, "will update prefs");
		$update = 1; 
	}
	# bigger increments for bigger numbers
	if ( $xPos{$id} == 2 ) {
		$amount *= -1;	# so up increases the numbers
		if ( ($yPos{$id} + (30 * $amount)) > 600 ) {
			$log->msg('menu',80, "use 60s increments");
			$amount *= 60;
		} elsif ( ($yPos{$id} + (15 * $amount)) > 300 ) {
			$log->msg('menu',80, "use 30s increments");
			$amount *= 30;
		} elsif ( ($yPos{$id} + (5 * $amount)) > 90 ) {
			$log->msg('menu',80, "use 15s increments");
			$amount *= 15;
		} elsif ( ($yPos{$id} + (1 * $amount)) > 20 ) {
			$log->msg('menu',80, "use 5s increments");
			$amount *= 5;
		}
		$maxPos = undef;
	} else {
		if ( scalar(@choices) == 0 ) {
			@choices = @{$menuItems[$xPos{$id}]};
		}
		$log->msg('menu',80, "choices: ".join(" ",@choices));
		$maxPos = scalar(@choices) - 1;
	}
	$yPos{$id} += $amount;
	if ( ($xPos{$id} == 2) && ($yPos{$id} > 0) && ($yPos{$id} < $minActiveDuration) ) { 
		if ( $amount > 0 ) {
			$log->msg('menu',80, "increasing to $minActiveDuration seconds");
			$yPos{$id} = $minActiveDuration;
		} else {
			$log->msg('menu',80, "decreasing to 0 seconds");
			$yPos{$id} = 0;
		}
	}
	if ( ($xPos{$id} == 2) && ($yPos{$id} < 0) ) {
		$log->msg('menu',80, "cannot decrease below 0 seconds");
		$update = 0;
		$yPos{$id} = 0;
	}
	if ( $yPos{$id} < 0 ) { 
		$log->msg('menu',80, "wrapping to $maxPos position");
		$yPos{$id} = $maxPos; 
	}
	if ( defined($maxPos) && ($yPos{$id} > $maxPos) ) { 
		$log->msg('menu',80, "wrapping to position 0");
		$yPos{$id} = 0; 
	}
	if ( $xPos{$id} == 1 ) {
		$log->msg('menu',50, "setting menuSaverNpos{$id} to $yPos{$id}");
		$menuSaverNpos{$id} = $yPos{$id};
		#$log->msg("saverNPos $menuSaverNpos{$id}\n");
	}
	# update X name hash for this position
	my $thisPos;
	if ( $xPos{$id} == 2 ) {
		$thisPos = CGI::escape($yPos{$id});
		$log->msg('menu',80, "thisPos $thisPos");
	} elsif ( ($xPos{$id} == 1) && ($xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL|) ) {
		$thisPos = CGI::escape($yPos{$id});
		$log->msg('menu',80, "thisPos $thisPos");
	} else {
		$thisPos = CGI::escape(${$menuItems[$xPos{$id}]}[$yPos{$id}]);
		$log->msg('menu',80, "thisPos $thisPos");
	}
	my @xArray = split(/\t/,$xHash{$id});
	$xArray[$xPos{$id}] = $thisPos;
	$xHash{$id} = join("\t",@xArray)."\t";
	# update prefs?
	if ( $update == 1 ) {
		$log->msg('menu',80, "updating prefs");
		&updatePref($id,$client);
	}
}

sub getSaverType {
	my $string = shift;
	if ( $string eq 'SETUP_SCREENSAVER' ) { return 'screensaver'; }
	if ( $string eq 'SETUP_OFFSAVER' ) { return 'offsaver'; }
	return 'idlesaver';
}

sub getSaverName {
	my $string = shift;
	for (my $i = 0; $i < scalar(@saverNames); ++$i) {
		if ( $saverStringNames[$i] eq $string) {
			return $saverNames[$i];
		}
	}
	# default to first pos ("None" saver)
	return $saverNames[0];
}

sub getSaverString {
	my $name = shift;
	&loadScreensaverList();
	for (my $i = 0; $i < scalar(@saverNames); ++$i) {
#$log->msg("look for $name at $i, $saverNames[$i]\n");
		if ( $saverNames[$i] eq $name) {
			return $saverStringNames[$i];
		}
	}
	# none!
	return undef;
}

sub getSaverPos {
	my $string = shift;
	my @choices = @{$menuItems[3]};
	for (my $i = 0; $i < scalar(@choices); ++$i) {
#$log->msg("look for $string at $i, $choices[$i]\n");
		if ( $choices[$i] eq $string ) { return $i; }
	}
	# default to first
	return 0;
}

sub inactiveString() {
	my $buttonControl = Slim::Utils::Prefs::get("$screensaverName-button-control");
	if ( $buttonControl == 0 ) { return 'DISABLED'; }
	return 'MANUAL_ADJECTIVE';
}

sub displayLines {
	my $client = shift;
	my $id = $client->id();
	&initMenuPos($id);
	my ($x, $y) = ($xPos{$id}, $yPos{$id});
	my $line1 = string('PLUGIN_SAVERSWITCHER');
	my $line2 = '';
	my $prefix = '';
	my $suffix = '';
	if ( $x == 1 ) {
		if ( $xHash{$id} =~ m|^PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL| ) {
			$prefix = string('PLUGIN_SAVERSWITCHER_SAVER_BUTTON_CONTROL').' ';
			$line2 = $buttonChoices[$y];
		} else {
			my @choices = @{$menuItems[$x]};
			my $saverType = &getMenuSaverType($id);
			my ($sptr, $dptr) = &getSavers($saverType,undef);
			my @durations = @{$dptr};
			my @savers = @{$sptr};
			$prefix = $choices[$y].' ';
			if ( defined($durations[$y]) && ($durations[$y] > 0) ) {
				$prefix .= &minSecString($durations[$y]).' ';
			} else {
				$prefix .= '('.string(&inactiveString()).') ';
			}
			$line2 = &getSaverString($savers[$y]);
		}
	} elsif ( $x == 2 ) {
		if ( $y > 0 ) {
			$prefix = "$y ";
			$line2 = 'SECONDS';
		} else {
			$line2 = &inactiveString();
		}
		# add " (m:ss)" if $y >= 60
		# we don't do " (h:mm:ss)" because durations > 1 hour are not expected
		if ( $y >= 60 ) {
			my $min = int($y/60);
			my $sec = $y % 60;
			$suffix = ' '.&minSecString($y);
		}
	} else {
		my @choices = @{$menuItems[$x]};
		$line2 = $choices[$y];
	}
	my $localizedString = string($line2);
	if ( defined($localizedString) && ($localizedString ne '') ) {
		$line2 = $localizedString;
	}
	$line2Hash{"$id-$x"} = $prefix.$line2;
	$line2 = $prefix.$line2.$suffix;
	if ( defined($line2Hash{"$id-".($x - 1)}) ) {
		$line1 .= ' '.$line2Hash{"$id-".($x - 1)};
	}
        return($line1,$line2);
}

sub minSecString($) {
	my $y = shift;
	my $min = int($y/60);
	my $sec = $y % 60;
	return sprintf("(%i:%-.2i)",$min,$sec);
}

sub initMenuPos {
	my $id = shift;
	if (! defined($xPos{$id}) ) {
		$xPos{$id} = 0;
		$yPos{$id} = 0;
		$menuSaverNpos{$id} = $yPos{$id};
		$xHash{$id} = '';
		$xHash{$id} .= CGI::escape(${$menuItems[$xPos{$id}]}[$yPos{$id}])."\t";
		$log->msg('menu',50, "$id: start at xPos 0, yPos 0, menuSaverNpos 0, xHash \"$xHash{$id}\"");
	}
	$log->msg('menu',80, "$id: xPos $xPos{$id}, yPos $yPos{$id}, menuSaverNpos $menuSaverNpos{$id}, xHash \"$xHash{$id}\"");
	&loadScreensaverList();
	$menuItems[3] = \@saverStringNames;
}

sub getFunctions {
        return \%menuFunctions;
}

sub loadScreensaverList {
	my %hash  = %{&Slim::Buttons::Common::hash_of_savers};
	my @savers;
	my $i = 1;
	@saverNames = ();
	@saverStringNames = ();
	foreach my $name (sort keys %hash) {
		my $stringName = $hash{$name};
		my $use = 1;
		#$log->msg("saver $i: name \"$name\" string name \"$stringName\"\n");
		foreach my $badsaver ( @incompatibleSavers ) {
			# check if it's incompatible, if so $use = 0
			if ( $name eq $badsaver ) {
				$use = 0;
			}
		}
		if ( $use == 1 ) {
			push @saverNames, $name;
			push @saverStringNames, $stringName;
			++$i;
		}
	}
	#$log->msg("$i screensavers recognized\n");
}

sub registerPreUsageCheck {
	my ($saverName, $coderef) = @_;
	if (! defined($coderef) ) {
		# the plugin is indicating no pre-usage check should be used
		$log->msg('funcs',50,"un-registering pre-change coderef for $saverName");
		$preChangeFunctions{$saverName} = undef;
		return 1;
	} 
	if ( ref($coderef) ne 'CODE' ) {
		$log->msg("Error: SaverSwitcher::Public::registerPreUsageCheck() called for $saverName without coderef!");
		return 0;
	}
	$log->msg('funcs',50,"registering pre-change coderef for $saverName");
	$preChangeFunctions{$saverName} = $coderef;
	return 1;
}

1;
